#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include <stdio.h>


#define ON			1
#define OFF			0

#define NONE		0
#define ADULT		1
#define KID			2

#define FORWARD		0
#define REWARD		1

// Key Define
#define KEY_PIN		PINE
#define KEY_DDR		DDRE

#define KEY1		1
#define KEY2		2
#define KEY3		3
#define KEY4		4
#define KEY5		5

// Lcd Define
#define LCD_PORT		PORTC
#define LCD_PIN		PINC
#define LCD_DDR		DDRC

#define LCD_RS		0x01
#define LCD_RW		0x02
#define LCD_E		0x04

#define DDRAM		0x80

// FND Define
#define FND_PORT		PORTB
#define FND_DDR		DDRB
#define FND_DIG0		0x04
#define FND_DIG1		0x08

// Matrix LED Define
#define MLED_PORT		PORTA
#define MLED_DDR		DDRA
#define MLED_DATA		0x01
#define MLED_SHIFT	0x02
#define MLED_CLEAR	0x04
#define MLED_LATCH1	0x10
#define MLED_LATCH2	0x20

// Serial Baudrate Define
#define B4800		143
#define B9600		71
#define B19200		35
#define B38400		17
#define B57600		11
#define B115200		5

enum st {
	SELECT_TICKET_TYPE,
	SELECT_START_STATION,
	SELECT_END_STATION,
	CALCULATE_FEE,
	TRAIN_GO
};


// Global Variable Define
volatile unsigned char key_flag;
volatile char fnd_buf[2], fnd_flag;
volatile int dig;
unsigned int led_order[16] = {
	0x0001, 0x0002, 0x0004, 0x0008, 0x0080, 0x0800, 0x8000, 0x4000,
	0x0400, 0x0040, 0x0020, 0x0200, 0x2000, 0x1000, 0x0100, 0x0010
};
volatile char state;
volatile char ticket_type, direction;
volatile int start_station, end_station, distance;
int fee_tbl[2] = { 300, 500 }, fee, money;


void timer0_init(void)
{
	TCCR0 = 0x07;
	TCNT0 = 256 - 11;
	TIMSK = 0x01;
}

SIGNAL(TIMER0_OVF_vect)
{
	TCNT0 = 256 - 11;

	if( fnd_flag ) {
		FND_PORT = 0x00;
		FND_PORT = (fnd_buf[dig] << 4) | (FND_DIG0 << dig);

		dig ^= 0x01;
	}
}

void fnd_off(void)
{
	fnd_flag = OFF;
	FND_PORT = 0x00;
}

void matrix_led(int data)
{
	int tmp = ~data;
	int i;

	MLED_PORT |= MLED_CLEAR;
	MLED_PORT &= ~(MLED_LATCH1 | MLED_LATCH2);
	for( i = 0; i < 16; i++ ) {
		if( tmp & 0x8000 )	MLED_PORT |= MLED_DATA;
		else				MLED_PORT &= ~MLED_DATA;

		MLED_PORT |= MLED_SHIFT;
		tmp <<= 1;
		MLED_PORT &= ~MLED_SHIFT;
	}

	MLED_PORT |= (MLED_LATCH1 | MLED_LATCH2);
	MLED_PORT &= ~(MLED_LATCH1 | MLED_LATCH2);
}

void matrix_alloff(void)
{
	int i;

	MLED_PORT |= MLED_CLEAR;
	MLED_PORT &= ~(MLED_LATCH1 | MLED_LATCH2);
	for( i = 0; i < 16; i++ ) {
		MLED_PORT |= MLED_DATA;

		MLED_PORT |= MLED_SHIFT;
		MLED_PORT &= ~MLED_SHIFT;
	}

	MLED_PORT |= (MLED_LATCH1 | MLED_LATCH2);
	MLED_PORT &= ~(MLED_LATCH1 | MLED_LATCH2);
}

void matrix_allon(void)
{
	MLED_PORT &= ~MLED_CLEAR;
	MLED_PORT |= (MLED_LATCH1 | MLED_LATCH2);
	MLED_PORT &= ~(MLED_LATCH1 | MLED_LATCH2);
}

char getkey(char prev_key)
{
	char key = 0, old_key;

	key_flag = 0;

	old_key = ~(KEY_PIN >> 3) & 0x1F;
	if( old_key ) {
		_delay_ms(50);
		key = ~(KEY_PIN >> 3) & 0x1F;
		if( key == old_key ) {
			if( key == 0x01 )		key = KEY1;
			else if( key == 0x02 )	key = KEY2;
			else if( key == 0x04 )	key = KEY3;
			else if( key == 0x08 )	key = KEY4;
			else if( key == 0x10 )	key = KEY5;

			if( key != prev_key )	key_flag = 1;
		}
	}

	return key;
}

void lcd_busycheck(void)
{
	LCD_DDR = 0x0F;

	LCD_PORT = 0x00;
	LCD_PORT |= LCD_RW;
	LCD_PORT |= LCD_E;
	while( LCD_PIN & 0x80 );
	LCD_PORT &= ~LCD_E;

	LCD_DDR = 0xFF;
}

unsigned char lcd_command_read(void)
{
	char cmd;

	LCD_DDR = 0x0F;
	LCD_PORT = 0x00;

	LCD_PORT |= LCD_RW;

	LCD_PORT |= LCD_E;
	asm	("nop");
	cmd = LCD_PIN & 0xF0;
	LCD_PORT &= ~LCD_E;

	LCD_PORT |= LCD_E;
	asm	("nop");
	cmd |= LCD_PIN >> 4;
	LCD_PORT &= ~LCD_E;

	return cmd;
}

void lcd_command_write(unsigned char cmd)
{
	lcd_busycheck();

	LCD_PORT = 0x00;

	LCD_PORT |= cmd & 0xF0;
	LCD_PORT |= LCD_E;
	LCD_PORT &= ~LCD_E;

	LCD_PORT &= 0x0F;
	LCD_PORT |= cmd << 4;
	LCD_PORT |= LCD_E;
	LCD_PORT &= ~LCD_E;

	_delay_ms(2);
}

unsigned char lcd_data_read(void)
{
	char data;

	LCD_DDR = 0x0F;
	LCD_PORT = 0x00;

	LCD_PORT |= LCD_RS;
	LCD_PORT |= LCD_RW;

	LCD_PORT |= LCD_E;
	asm	("nop");
	data = LCD_PIN & 0xF0;
	LCD_PORT &= ~LCD_E;

	LCD_PORT |= LCD_E;
	asm	("nop");
	data |= LCD_PIN >> 4;
	LCD_PORT &= ~LCD_E;

	return data;
}

void lcd_data_write(unsigned char data)
{
	LCD_PORT = 0x00;

	LCD_PORT |= LCD_RS;

	LCD_PORT |= data & 0xF0;
	LCD_PORT |= LCD_E;
	LCD_PORT &= ~LCD_E;

	LCD_PORT &= 0x0F;
	LCD_PORT |= data << 4;
	LCD_PORT |= LCD_E;
	LCD_PORT &= ~LCD_E;

	_delay_us(50);
}

void lcd_string(char *str)
{
	while( *str )	lcd_data_write(*str++);
}

void lcd_gotoxy(char x, char y)
{
	lcd_command_write(DDRAM | (0x40 * y) | x);
}

void lcd_init(void)
{
	LCD_DDR = 0xFF;

	lcd_command_write(0x20);
	_delay_ms(10);
	lcd_command_write(0x20);
	_delay_ms(10);
	lcd_command_write(0x20);
	_delay_ms(10);
	lcd_command_write(0x28);
	lcd_command_write(0x08);
	lcd_command_write(0x01);
	lcd_command_write(0x06);
	_delay_ms(10);
	lcd_command_write(0x0C);
}

void uart_transmit(unsigned char data)
{
	while( !( UCSR1A & (1 << UDRE1)) );

	UDR1 = data;
}

void uart_string(char *str)
{
	while( *str )	uart_transmit(*str++);
}

unsigned char uart_receive(void)
{
	while ( !(UCSR1A & (1 << RXC)) );

	return UDR1;
}

void uart_init(unsigned char baudrate)
{
	UCSR1A = (0 << RXC) | (0 << TXC) | (0 << UDRE) | (0 << FE) | (0 << DOR) | (0 << UPE) | (0 << U2X) | (0 << MPCM);
	UCSR1B = (1 << RXCIE) | (0 << TXCIE) | (0 << UDRIE) | (1 << RXEN) | (1 << TXEN) | (0 << UCSZ2) | (0 << RXB8) | (0 << TXB8);
	UCSR1C = (0 << UMSEL) | (0 << UPM1) | (0 << UPM0) | (0 << USBS) | (1 << UCSZ1) | (1 << UCSZ0) | (0 << UCPOL);
	UBRR1H = 0x00;
	UBRR1L = baudrate;
}

void adc_init(void)
{
	ADCSRA = 0b10001111;
	ADMUX = 0b00000000;
}

void adc_start(void)
{
	ADCSRA |= 0x40;
}

SIGNAL(ADC_vect)
{
	money = (int)((1023 - ADC) / 20.46) * 100;
	ADCSRA |= 0x40;
}

void main_screen(void)
{
	fnd_off();
	matrix_alloff();

	lcd_gotoxy(0, 0);
	lcd_string(" Select Button! ");
	lcd_gotoxy(0, 1);
	lcd_string("S1:Adult  S2:Kid");
}

void start_screen(void)
{
	int i;

	lcd_string("   Welcome To   ");
	lcd_gotoxy(0, 1);
	lcd_string("Gangnam Station!");

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	[ 1]	LED  
//			. LED :   0.3  ϰ Ͻÿ.
//				LED1LED2LED3LED4LED8LED12LED16LED15
//				LED11LED7LED6LED10LED14LED13LED9LED5
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
								// LED 16 ̹Ƿ 16ȸ ݺ
								// matrix_led Լ ̿Ͽ LED ON
								// 0.3 


	main_screen();
}

void mcu_init(void)
{
	KEY_DDR = 0x00;
	FND_DDR = 0xFC;
	FND_PORT = 0xF0;
	MLED_DDR = 0xFF;
	MLED_PORT = 0x00;

	adc_init();
	lcd_init();
	uart_init(B9600);
	timer0_init();

	sei();
}

void variable_init(void)
{
	key_flag = OFF;

	dig = 0;
	fnd_off();

	ticket_type = NONE;
	start_station = NONE;
	end_station = NONE;

	distance = 0;
	fee = 0;
}

void select_station(char keyin)
{
	char key = keyin;
	int station;

	_delay_ms(500);

	start_station = 1;
	station = start_station;

	distance = 0;
	fee = 0;

	lcd_gotoxy(0, 0);
	lcd_string(" Start Station  ");
	lcd_gotoxy(0, 1);
	lcd_string("Station Num : ");

	while( 1 ) {
		key = getkey(key);
		if( key_flag ) {
			switch( key ) {
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	[ 2] ġ ̿Ͽ LED  Ǵ  ̵
//		. LED
//			SW1 1   1徿  ̵ǰ,
//			SW2 1   1徿  ̵ǰ Ͻÿ.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
				case KEY1 :			//  1  ̵(1 )


					break;

				case KEY2 :			//  1  ̵(1 )


					break;

				case KEY3 :
					if( state == SELECT_START_STATION ) {
						start_station = station;
						end_station = 1;
						station = end_station;

						lcd_gotoxy(0, 0);
						lcd_string(" End   Station  ");

						state = SELECT_END_STATION;
					}
					else if( state == SELECT_END_STATION ) {
						end_station = station;

						lcd_gotoxy(0, 0);
						printf("Start:%02d Station", start_station);
						lcd_gotoxy(0, 1);
						printf("End  :%02d Station", end_station);

						matrix_led(led_order[start_station - 1] | led_order[end_station - 1]);

						_delay_ms(500);

						if( start_station == end_station ) {
							lcd_gotoxy(0, 0);
							lcd_string("    Error!!     ");
							lcd_gotoxy(0, 1);
							lcd_string("Return To First!");
							_delay_ms(500);

							start_station = NONE;
							end_station = NONE;

							main_screen();

							state = SELECT_TICKET_TYPE;
						}
						else {
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	[ 3] ߿    ִܰŸ  Ͽ LCD ǥ
//		. LCD
//			Ÿ : ߿   ִ Ÿ Ͽ ǥ
//			 : Adult = 500 X Ÿ, Kid = 300 X Ÿ
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
							distance = end_station - start_station;
							if( distance < 0 ) {
								distance *= (-1);
								if( distance < 8 ) {
									direction = REWARD;
								}
								else {
									direction = FORWARD;
									distance = 16 - distance;
								}
							}
							else {
								if( distance > 8 ) {
									direction = REWARD;
									distance = 16 - distance;
								}
								else {
									direction = FORWARD;
								}
							}

							// LCD ù° ٿ ִܰŸ ǥ



							// LCD ι° ٿ  Ͽ ǥ




							_delay_ms(1000);

							lcd_gotoxy(0, 0);
							lcd_string("Insert Money!!  ");
							lcd_gotoxy(0, 1);
							printf("Fee:%04d-0000   ", fee);

							state = CALCULATE_FEE;
						}

						return;
					}
					break;

				case KEY5 :
					wdt_enable(WDTO_30MS);
					while( 1 );
					break;
			}
		}

		matrix_led(led_order[station - 1]);
		lcd_gotoxy(14, 1);
		printf("%02d", station);
	}
}

void train_go(void)
{
	int i, pos = start_station - 1;
	char str[30];

	lcd_gotoxy(0, 0);
	lcd_string("Train Is Going..");

	fnd_flag = ON;
	for( i = 0; i <= distance; i++ ) {
		matrix_led(led_order[pos]);

		if( i < distance ) {
			lcd_gotoxy(0, 1);
			if( direction == FORWARD )	printf("----> : %d Minutes", distance - i);
			else					printf("<---- : %d Minutes", distance - i);
		}
		else {
			lcd_gotoxy(0, 0);
			lcd_string("Train Is Arrived");
			lcd_gotoxy(0, 1);
			lcd_string(" Thanks A Lot!  ");
		}

		fnd_buf[0] = i;
		fnd_buf[1] = distance - i;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	[ 5] 
//		 Ŀ 缭 ǽð PC ͹̳ο Ͻÿ.
//		̵ð, Ÿ, ߿, , 翪,  ͹̳ο 
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		uart_transmit('\f');
		//	̵ð 


		//	 Ÿ 


		//	߿ 


		//	 


		//	翪 


		//	 



		if( direction == FORWARD ) {
			if( pos < 15 )	pos++;
			else			pos = 0;
		}
		else {
			if( pos )	pos--;
			else		pos = 15;
		}

		_delay_ms(1000);
	}
	_delay_ms(1000);

	ticket_type = NONE;
	start_station = NONE;
	end_station = NONE;

	distance = 0;
	fee = 0;

	main_screen();

	state = SELECT_TICKET_TYPE;
}

int main(void)
{
	char key = 0xFF;

	mcu_init();
	variable_init();
	matrix_alloff();

	adc_start();

	fdevopen((void *)lcd_data_write, 0);

	start_screen();

	state = SELECT_TICKET_TYPE;

	while( 1 ) {
		key = getkey(key);
		if( key_flag ) {
			switch( key ) {
				case KEY1 :
				case KEY2 :
					if( state == SELECT_TICKET_TYPE ) {
						state = SELECT_START_STATION;

						lcd_command_write(0x01);
						lcd_gotoxy(0, 0);
						if( key == KEY1 )		lcd_string("You Select Adult");
						else if( key == KEY2 )	lcd_string("You Select Kid  ");

						ticket_type = key;
						select_station(key);
					}
					break;

				case KEY3 :
					if( state == CALCULATE_FEE ) {
						if( money >= fee ) {
							if( money > fee ) {
								lcd_gotoxy(0, 0);
								lcd_string("Here Is Change. ");
								lcd_gotoxy(0, 1);
								printf("Change : %04d   ", money - fee);
							}
							_delay_ms(1000);

							lcd_gotoxy(0, 0);
							lcd_string("  [S4] Go!!     ");
							lcd_gotoxy(0, 1);
							printf("  Distance : %d  ", distance);

							state = TRAIN_GO;
						}
					}
					break;

				case KEY4 :
					if( state == TRAIN_GO )	train_go();
					break;

				case KEY5 :
					wdt_enable(WDTO_30MS);
					while( 1 );
					break;
			}
		}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	[ 4]  ̿Ͽ    ǥ
//		1. VR1 ADC Է°  ݾ LCD ǥä
//		2. Էµ ݾ ݰ Ͽ ũų  'OK' ǥϰ
//			   ǰ Ͻÿ.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		if( state == CALCULATE_FEE ) {



			if( money >= fee ) {


			}
		}
	}

	return 0;
}
